import os.path
import sys

from glad.lang.common.generator import Generator
from glad.lang.common.util import makefiledir

if sys.version_info >= (3, 0):
    basestring = str


def _gl_types(gen, f):
    f.write('''
  GLdebugProc = procedure (
    source: GLenum;
    typ: GLenum;
    id: GLuint;
    severity: GLenum;
    length: GLsizei;
    message: PGLchar;
    userParam: pointer); stdcall;
  GLdebugProcArb = GLdebugProc;
  GLdebugProcKhr = GLdebugProc;

  GLdebugProcAmd = procedure (
    id: GLuint;
    category: GLenum;
    severity: GLenum;
    len: GLsizei;
    message: PGLchar;
    userParam: pointer); stdcall;
''')

# TODO egl, glx & wgl loaders
# def _egl_types(gen, f):
# def _glx_types(gen, f):
# def _wgl_types(gen, f):

PASCALTYPES = {
    '__other': {
        'gl': _gl_types
        # 'egl': _egl_types,
        # 'glx': _glx_types,
        # 'wgl': _wgl_types
    },

    'gl': {
        'GLbitfield': 'uint32',
        'GLboolean': 'byte',
        'GLbyte': 'int8',
        'GLchar': 'char',
        'GLcharARB': 'byte',
        'GLclampd': 'double',
        'GLclampf': 'single',
        'GLclampx': 'int32',
        'GLdouble': 'double',
        'GLeglImageOES': 'pointer',
        'GLenum': 'uint32',
        'GLfixed': 'int32',
        'GLfloat': 'single',
        'GLhalf': 'uint16',
        'GLhalfARB': 'uint16',
        'GLhalfNV': 'uint16',
        'GLhandleARB': 'uint32',
        'GLint': 'int32',
        'GLint64': 'int64',
        'GLint64EXT': 'int64',
        'GLintptr': 'int32',
        'GLintptrARB': 'int32',
        'GLshort': 'int16',
        'GLsizei': 'int32',
        'GLsizeiptr': 'int32',
        'GLsizeiptrARB': 'int32',
        'GLubyte': 'uint8',
        'GLuint': 'uint32',
        'GLuint64': 'uint64',
        'GLuint64EXT': 'uint64',
        'GLushort': 'uint16',
        'GLvdpauSurfaceNV': 'int32',
        'GLvoid': 'pointer',
        'GLsync': 'pointer',
        'GLeglClientBufferEXT': 'pointer',  # GL_EXT_external_buffer
        'GLVULKANPROCNV': 'pointer',  # GL_NV_draw_vulkan_image
        '_cl_context': 'pointer',  # GL_ARB_cl_event
        '_cl_event': 'pointer'
    },
    'egl': {
        'EGLAttrib': 'int32',
        'EGLAttribKHR': 'int32',
        'EGLBoolean': 'dword',
        'EGLClientBuffer': 'pointer',
        'EGLConfig': 'pointer',
        'EGLContext': 'pointer',
        'EGLDeviceEXT': 'pointer',
        'EGLDisplay': 'pointer',
        'EGLImage': 'pointer',
        'EGLImageKHR': 'pointer',
        'EGLNativeFileDescriptorKHR': 'int32',
        'EGLOutputLayerEXT': 'pointer',
        'EGLOutputPortEXT': 'pointer',
        'EGLStreamKHR': 'pointer',
        'EGLSurface': 'pointer',
        'EGLSync': 'pointer',
        'EGLSyncKHR': 'pointer',
        'EGLSyncNV': 'pointer',
        'EGLTimeKHR': 'uint64',
        'EGLTime': 'uint64',
        'EGLTimeNV': 'uint64',
        'EGLenum': 'uint32',
        'EGLint': 'int32',
        'EGLsizeiANDROID': 'pointer',
        'EGLuint64KHR': 'uint64',
        'EGLuint64MESA': 'uint64',
        'EGLuint64NV': 'uint64',
        #        '__eglMustCastToProperFunctionPointerType': 'void function()'
    },
    'glx': {
        'GLbitfield': 'uint32',
        'GLboolean': 'uint8',
        'GLenum': 'uint32',
        'GLfloat': 'float32',
        'GLint': 'int32',
        'GLintptr': 'int32',
        'GLsizei': 'int32',
        'GLsizeiptr': 'int32',
        'GLubyte': 'uint8',
        'GLuint': 'uint32'
    },
    'wgl': {
        'GLbitfield': 'uint32',
        'GLboolean': 'uint8',
        'GLenum': 'uint32',
        'GLfloat': 'float32',
        'GLint': 'int32',
        'GLsizei': 'int32',
        'GLuint': 'uint32',
        'GLushort': 'uint16',
        'INT32': 'int32',
        'INT64': 'int64'
    },

    'SpecialNumbers': {
        'gl': [
            ('GL_FALSE', '0', None),
            ('GL_INVALID_INDEX', '$FFFFFFFF', 'uint32'),
            ('GL_NONE', '0', None),
            ('GL_NONE_OES', '0', None),
            ('GL_NO_ERROR', '0', None),
            ('GL_ONE', '1', None),
            ('GL_TIMEOUT_IGNORED', '$FFFFFFFFFFFFFFFF', 'uint64'),
            ('GL_TIMEOUT_IGNORED_APPLE', '$FFFFFFFFFFFFFFFF', 'uint64'),
            ('GL_TRUE', '1', None),
            ('GL_VERSION_ES_CL_1_0', '1', None),
            ('GL_VERSION_ES_CL_1_1', '1', None),
            ('GL_VERSION_ES_CM_1_1', '1', None),
            ('GL_ZERO', '0', None),
        ],
        'egl': [
        ],
        'glx': [
        ],
        'wgl': [
        ]
    }
}


class PascalGenerator(Generator):
    NAME = 'pascal'
    NAME_LONG = 'Pascal'

    MODULE = 'glad'
    FILE_EXTENSION = '.pas'
    TYPE_DICT = PASCALTYPES

    LOAD_GL_PREFIX = 'glad'
    EXT_PREFIX = 'GLAD_'

    def open(self):
        self._f_gl = open(self.make_path('glad_' + self.spec.NAME), 'w')

    def close(self):
        self._f_gl.close()

    def generate_header(self):
        f = self._f_gl
        f.write('{\n')
        f.write(self.header)
        f.write('}\n')
        f.write('''{$MODE objfpc}{$H+}
{$MACRO ON}
{$IFDEF Windows}
  {$DEFINE extdecl := stdcall}
{$ELSE}
  {$DEFINE extdecl := cdecl}
{$ENDIF}
''')
        f.write('unit glad_gl;\n\n')
        f.write('interface\n\n')
        f.write('uses\n  sysutils;\n\n')
        f.write('var\n  glVersionMajor, glVersionMinor: integer;\n\n')

    def generate_loader(self, features, extensions):
        f = self._f_gl

        # finish interface
        f.write('type\n  TLoadProc = function(proc: Pchar): Pointer;\n\n')
        loadername = 'Load' if self.LOAD_GL_PREFIX else 'load'
        loader_decl = 'function {}{}{}(load: TLoadProc): boolean;\n'
        for api, version in self.api.items():
            decl = loader_decl.format(self.LOAD_GL_PREFIX, loadername, api.upper())
            f.write(decl)
        f.write('\n\nimplementation\n\n')

        self.loader.write(f)
        self.loader.write_has_ext(f)

        written = set()
        for api, version in self.api.items():
            # core load procs
            for feature in features[api]:
                f.write('procedure load_{}(load: TLoadProc);\nbegin\n'.format(feature.name))
                if self.spec.NAME == 'gl':
                    f.write('  if not {}{} then exit;\n'.format(self.EXT_PREFIX, feature.name))
                for func in feature.functions:
                    self.write_func_definition(f, func)
                f.write('end;\n\n')

            # extension load procs
            for ext in extensions[api]:
                if len(list(ext.functions)) == 0 or ext.name in written:
                    continue

                f.write('procedure load_{}(load: TLoadProc);\nbegin\n'.format(ext.name))
                if self.spec.NAME == 'gl':
                    f.write('  if not {}{} then exit;\n'.format(self.EXT_PREFIX, ext.name))
                for func in ext.functions:
                    # even if they were in written we need to load it
                    self.write_func_definition(f, func)
                f.write('end;\n\n')

                written.add(ext.name)

            # findExtensions proc
            f.write('procedure findExtensions{}();\nbegin\n'.format(api.upper()))
            if self.spec.NAME == 'gl':
                for ext in extensions[api]:
                    f.write('  {0}{1} := hasExt(\'{1}\');\n'.format(self.EXT_PREFIX, ext.name))
            f.write('end;\n\n')

            # findCore proc
            f.write('procedure findCore{}(glVersion: string);\n'.format(api.upper()))
            self.loader.write_find_core(f)
            if self.spec.NAME == 'gl':
                for feature in features[api]:
                    f.write('  {}{} := ((major = {num[0]}) and (minor >= {num[1]})) or'
                            ' (major > {num[0]});\n'.format(self.EXT_PREFIX, feature.name,
                                                            num=feature.number))
            f.write('end;\n\n')

            # main loader proc
            decl = loader_decl.format(self.LOAD_GL_PREFIX, loadername, api.upper())
            f.write(decl)
            self.loader.write_begin_load(f)

            f.write('  findCore{}(glVersion);\n'.format(api.upper()))
            for feature in features[api]:
                f.write('  load_{}(load);\n'.format(feature.name))
            f.write('\n  findExtensions{}();\n'.format(api.upper()))
            for ext in extensions[api]:
                if len(list(ext.functions)) == 0:
                    continue
                f.write('  load_{}(load);\n'.format(ext.name))
            self.loader.write_end_load(f)
            f.write('\n')

        f.write('end.\n')

    def write_func_definition(self, fobj, func):
        func_name = self.map_func_name(func)
        fobj.write('  pointer( {0} ) := load(\'{0}\');\n'.format(func_name))

    def map_func_name(self, func):
        name = func.proto.name
        return name

    def generate_types(self, types):
        f = self._f_gl

        f.write('(* Types *)\ntype\n')
        for ogl, pascal in sorted(self.TYPE_DICT[self.spec.NAME].items()):
            f.write('  {} = {};\n'.format(ogl, pascal))
        f.write('\n')

        # pointer types
        for ogl, pascal in sorted(self.TYPE_DICT[self.spec.NAME].items()):
            f.write('  P{} = ^{};\n'.format(ogl, pascal))
        f.write('  PPGLchar = ^PGLchar;\n')
        f.write('  PPGLcharARB = ^PGLcharARB;\n')  # for GL_ARB_shader_objects
        f.write('  PPGLboolean = ^PGLboolean;\n')  # for GL_IBM_vertex_array_lists
        self.TYPE_DICT['__other'][self.spec.NAME](self, f)
        f.write('\n')

    def generate_features(self, features):
        self.write_enums(features)
        self.write_funcs(features)

    def write_enums(self, features):
        f = self._f_gl

        f.write('\n(* Enums *)\nconst\n')
        for v in sorted(self.TYPE_DICT['SpecialNumbers'][self.spec.NAME]):
            self.write_enum(f, *v)
        f.write('\n')

        written = set()
        for feature in features:
            for enum in feature.enums:
                if enum.group == 'SpecialNumbers':
                    written.add(enum)
                    continue
                if not enum in written:
                    self.write_enum(f, enum.name, enum.value)
                written.add(enum)
        f.write('\n')

    def write_funcs(self, features):
        f = self._f_gl

        f.write('(* Functions *)\nvar\n')
        if self.spec.NAME == 'gl':
            for feature in features:
                self.write_boolean(f, feature.name)
            f.write('\n')
            self.write_functions(f, set(), set(), features)
            # TODO other specs
        f.write('\n\n')

    def generate_extensions(self, extensions, enums, functions):
        if not extensions:
            return

        write = set()
        written = set(enum.name for enum in enums) | \
                  set(function.proto.name for function in functions)

        f = self._f_gl
        f.write('(* Extensions *)\n')

        for ext in extensions:
            if self.spec.NAME == 'gl' and ext.name not in written:
                f.write('var\n')
                self.write_boolean(f, ext.name)

            first = True
            for enum in ext.enums:
                if enum.name not in written and not enum.group == 'SpecialNumbers':
                    # for NV_transform_feedback - these enums are negative, but GLenum is unsigned
                    type = None if enum.group == 'TransformFeedbackTokenNV' else 'GLenum'
                    if first:
                        first = False
                        f.write('const\n')
                    self.write_enum(f, enum.name, enum.value, type)
                written.add(enum.name)
            written.add(ext.name)

        f.write('\n')
        self.write_functions(f, write, written, extensions)
        f.write('\n')

    def write_functions(self, f, write, written, extensions):
        first = True
        for ext in extensions:
            for func in ext.functions:
                if func.proto.name not in written:
                    if first:
                        first = False
                        f.write('var\n')
                    self.write_function_var(f, func)
                    write.add(func)
                written.add(func.proto.name)

    def write_function_var(self, fobj, func):
        fobj.write('  {}: '.format(self.map_func_name(func)))
        self.write_function_declaration(fobj, func)
        fobj.write('\n')

    def write_function_declaration(self, fobj, func):
        ret = func.proto.ret.to_pascal()
        is_func = ret != 'void'
        if is_func:
            fobj.write('function (')
        else:
            fobj.write('procedure (')
        fobj.write('; '.join('{}: {}'.format(self.to_pascal_param_name(param.name),
                                             param.type.to_pascal())
                             for param in func.params))
        fobj.write(')')
        if is_func:
            fobj.write(': {}'.format(ret))
        fobj.write('; extdecl;')

    PASCAL_KEYWORDS = [  # conflicting keywords only
        'array', 'end', 'in', 'label', 'object', 'out', 'packed', 'program', 'string', 'type', 'unit'
    ]

    def to_pascal_param_name(self, name):
        return '{}_'.format(name) if name in self.PASCAL_KEYWORDS else name

    def make_path(self, name):
        path = os.path.join(self.path, self.MODULE.split('.')[-1],
                            name.split('.')[-1] + self.FILE_EXTENSION)
        makefiledir(path)
        return path

    def write_boolean(self, fobj, name):
        fobj.write('  {}{}: boolean;\n'.format(self.EXT_PREFIX, name))

    def write_enum(self, fobj, name, value, type='GLenum'):
        fobj.write('  ' + name)
        if value[0:2] == '0x':
            value = '$' + value[2:]
        if type and type != 'GLenum':
            fobj.write(' = {}({});'.format(type, value))
        else:
            fobj.write(' = {};'.format(value))
        fobj.write('\n')




LOAD_OPENGL_DLL = '''
private global Library libGL;
extern(System) private alias gladGetProcAddressPtrType = void* function(const(char)*);
private global gladGetProcAddressPtrType gladGetProcAddressPtr;

%(pre)s
bool %(init)s() {
    version(Windows) {
        libGL = Library.load("opengl32.dll");
    } else version(OSX) {
        libGL = Library.loads([
            "../Frameworks/OpenGL.framework/OpenGL",
            "/Library/Frameworks/OpenGL.framework/OpenGL",
            "/System/Library/Frameworks/OpenGL.framework/OpenGL",
            "/System/Library/Frameworks/OpenGL.framework/Versions/Current/OpenGL"
        ]);
    } else {
        libGL = Library.loads(["libGL.so.1", "libGL.so"]);
    }

    if(libGL !is null) {
        version(Windows) {
            string sym = "wglGetProcAddress";
        } else {
            string sym = "glXGetProcAddressARB";
        }
        // returns null on OSX, but that's fine
        gladGetProcAddressPtr = cast(typeof(gladGetProcAddressPtr))libGL.symbol(sym);
        return true;
    }

    return false;
}

private struct StructToDg {
    void* instance;
    void* func;
}

%(pre)s
void* %(proc)s(string name) {
    if(libGL is null) return null;
    void* result;

    if(gladGetProcAddressPtr !is null) {
        // TODO: name.ptr
        result = gladGetProcAddressPtr(name.ptr);
    }
    if(result is null) {
        result = libGL.symbol(name);
    }

    return result;
}

%(pre)s
void %(terminate)s() {
    if(libGL !is null) {
        libGL.free();
    }
    return;
}
'''


from glad.lang.volt.loader.egl import EGLVoltLoader
from glad.lang.volt.loader.gl import OpenGLVoltLoader
from glad.lang.volt.loader.glx import GLXVoltLoader
from glad.lang.volt.loader.wgl import WGLVoltLoader

from glad.lang.volt.generator import VoltGenerator


_specs = {
    'egl': EGLVoltLoader,
    'gl': OpenGLVoltLoader,
    'glx': GLXVoltLoader,
    'wgl': WGLVoltLoader
}

_generators = {
    'volt': VoltGenerator,
}


def get_generator(name, spec):
    gen = _generators.get(name)
    loader = _specs.get(spec)

    return gen, loader



from glad.lang.common.loader import BaseLoader
from glad.lang.nim.loader.glx import _GLX_LOADER


# TODO this is just a quick initial conversion of the D loader

_WGL_LOADER = _GLX_LOADER.replace('GLX', 'WGL')

_WGL_HAS_EXT = '''
private bool has_ext(const(char)* name) @nogc {
    return true;
}
'''


class WGLNimLoader(BaseLoader):
    def write_header_end(self, fobj):
        pass

    def write_header(self, fobj):
        pass

    def write(self, fobj):
        fobj.write('alias Loader = void* delegate(const(char)*);\n')
        if not self.disabled:
            fobj.write(_WGL_LOADER)

    def write_begin_load(self, fobj):
        pass

    def write_end_load(self, fobj):
        fobj.write('  return true\n')

    def write_find_core(self, fobj):
        pass

    def write_has_ext(self, fobj):
        fobj.write(_WGL_HAS_EXT)



class BaseLoader(object):
    def __init__(self, apis, disabled=False, local_files=False):
        self.apis = apis
        self.disabled = disabled
        self.local_files = local_files

    def write(self, fobj):
        raise NotImplementedError

    def write_begin_load(self, fobj):
        raise NotImplementedError

    def write_end_load(self, fobj):
        raise NotImplementedError

    def write_find_core(self, fobj):
        raise NotImplementedError

    def write_has_ext(self, fobj):
        raise NotImplementedError

    def write_header(self, fobj):
        raise NotImplementedError

    def write_header_end(self, fobj):
        raise NotImplementedError


class NullLoader(BaseLoader):
    def write_begin_load(self, fobj):
        pass

    def write_end_load(self, fobj):
        pass

    def write_header_end(self, fobj):
        pass

    def write_has_ext(self, fobj):
        pass

    def write(self, fobj):
        pass

    def write_header(self, fobj):
        pass

    def write_find_core(self, fobj):
        pass

    def __getattr__(self, name):
        try:
            return self.__getattribute__(name)
        except AttributeError:
            pass

        def dummy(*args, **kwargs):
            pass
        return dummy




from glad.lang.common.loader import BaseLoader


# TODO this is just a quick initial conversion of the D loader

_EGL_LOADER = '''
bool gladLoadEGL() {
    return gladLoadEGL(x => eglGetProcAddress(x))
}
'''

_EGL_HAS_EXT = '''
private bool has_ext(const(char)* ext) @nogc {
    return true
}
'''


class EGLNimLoader(BaseLoader):
    def write_header_end(self, fobj):
        pass

    def write_header(self, fobj):
        pass

    def write(self, fobj):
        if not self.disabled:
            fobj.write(_EGL_LOADER)

    def write_begin_load(self, fobj):
        pass

    def write_end_load(self, fobj):
        fobj.write('  return true\n')

    def write_find_core(self, fobj):
        pass

    def write_has_ext(self, fobj):
        fobj.write(_EGL_HAS_EXT)



from glad.lang.common.loader import BaseLoader
from glad.lang.nim.loader import LOAD_OPENGL_DLL


# TODO this is just a quick initial conversion of the D loader

_GLX_LOADER = \
    LOAD_OPENGL_DLL % {'pre':'private', 'init':'open_gl',
                       'proc':'get_proc', 'terminate':'close_gl'} + '''
bool gladLoadGLX() {
    bool status = false

    if(open_gl()) {
        status = gladLoadGLX(x => get_proc(x))
        close_gl()
    }

    return status
}
'''

_GLX_HAS_EXT = '''
private bool has_ext(const(char)* name) @nogc {
    return true
}
'''


class GLXNimLoader(BaseLoader):
    def write_header_end(self, fobj):
        pass

    def write_header(self, fobj):
        pass

    def write(self, fobj):
        fobj.write('alias Loader = void* delegate(const(char)*)\n')
        if not self.disabled:
            fobj.write(_GLX_LOADER)

    def write_begin_load(self, fobj):
        pass

    def write_end_load(self, fobj):
        fobj.write('  return true\n')

    def write_find_core(self, fobj):
        pass

    def write_has_ext(self, fobj):
        fobj.write(_GLX_HAS_EXT)



from itertools import chain
import os.path
import sys

from glad.lang.common.generator import Generator
from glad.lang.common.util import makefiledir


if sys.version_info >= (3, 0):
    from io import StringIO
    basestring = str
else:
    from StringIO import StringIO


def _gl_types(gen, f):
    gen.write_opaque_struct(f, '__GLsync')
    gen.write_alias(f, 'GLsync', '__GLsync*')
    gen.write_opaque_struct(f, '_cl_context')
    gen.write_opaque_struct(f, '_cl_event')
    gen.write_extern(f)
    gen.write_alias(
        f,
        'GLDEBUGPROC', 'void function(GLenum, GLenum, '
        'GLuint, GLenum, GLsizei, in GLchar*, GLvoid*)'
    )
    gen.write_alias(f, 'GLDEBUGPROCARB', 'GLDEBUGPROC')
    gen.write_alias(f, 'GLDEBUGPROCKHR', 'GLDEBUGPROC')
    gen.write_alias(
        f,
        'GLDEBUGPROCAMD', 'void function(GLuint, GLenum, '
        'GLenum, GLsizei, in GLchar*, GLvoid*)'
    )
    gen.write_extern_end(f)


def _egl_types(gen, f):
    io = StringIO()
    gen.write_opaque_struct(io, 'egl_native_pixmap_t')

    f.write('''
// Thanks to @jpf91 (github) for these declarations
version(Windows) {
    import core.sys.windows.windows;
    alias EGLNativeDisplayType = HDC;
    alias EGLNativePixmapType = HBITMAP;
    alias EGLNativeWindowType = HWND;
} else version(Symbian) {
    alias EGLNativeDisplayType = int;
    alias EGLNativeWindowType = void*;
    alias EGLNativePixmapType = void*;
} else version(Android) {
    //import android.native_window;
    //struct egl_native_pixmap_t;
    ''' + io.getvalue() + '''
    //alias ANativeWindow*           EGLNativeWindowType;
    //alias egl_native_pixmap_t*     EGLNativePixmapType;
    alias EGLNativeWindowType = void*;
    alias EGLNativePixmapType = void*;
    alias EGLNativeDisplayType = void*;
} else version(linux) {
    version(Xlib) {
        import X11.Xlib;
        import X11.Xutil;
        alias EGLNativeDisplayType = Display*;
        alias EGLNativePixmapType = Pixmap;
        alias EGLNativeWindowType = Window;
    } else {
        alias EGLNativeDisplayType = void*;
        alias EGLNativePixmapType = uint;
        alias EGLNativeWindowType = uint;
    }
}
alias EGLObjectKHR = void*;
alias EGLLabelKHR = void*;

extern(System) {
alias EGLSetBlobFuncANDROID = void function(const(void)*, EGLsizeiANDROID, const(void)*, EGLsizeiANDROID);
alias EGLGetBlobFuncANDROID = EGLsizeiANDROID function(const(void)*, EGLsizeiANDROID, const(void)* EGLsizeiANDROID);
struct EGLClientPixmapHI {
    void  *pData;
    EGLint iWidth;
    EGLint iHeight;
    EGLint iStride;
}
alias EGLDEBUGPROCKHR = void function(EGLenum error,const char *command,EGLint messageType,EGLLabelKHR threadLabel,EGLLabelKHR objectLabel,const char* message);
}
''')
    gen.write_extern(f)
    gen.write_opaque_struct(f, '_cl_event')
    gen.write_extern_end(f)


def _glx_types(gen, f):
    f.write('''
version(Xlib) {
    import X11.Xlib;
    import X11.Xutil;
} else {
    alias Bool = int;
    alias Status = int;
    alias VisualID = uint;
    alias XPointer = byte*;
    alias XID = uint;
    alias Colormap = XID;
    alias Display = void;
    alias Font = XID;
    alias Window = XID;
    alias Drawable = XID;
    alias Pixmap = XID;
    alias Cursor = XID;
    alias GContext = XID;
    alias KeySym = XID;

    extern(System) {
        // Borrowed from derelict
        struct XExtData {
            int number;
            XExtData* next;
            extern(C) int function(XExtData*) free_private;
            XPointer private_data;
        }

        struct Visual {
            XExtData* ext_data;
            VisualID  visualid;
            int       _class;
            uint      red_mask, green_mask, blue_mask;
            int       bits_per_rgb;
            int       map_entries;
        }

        struct XVisualInfo {
            Visual   *visual;
            VisualID visualid;
            int      screen;
            int      depth;
            int      _class;
            uint     red_mask;
            uint     green_mask;
            uint     blue_mask;
            int      colormap_size;
            int      bits_per_rgb;
        }
    }
}

alias DMbuffer = void*;
alias DMparams = void*;
alias VLNode = void*;
alias VLPath = void*;
alias VLServer = void*;

alias int64_t = long;
alias uint64_t = ulong;
alias int32_t = int;

alias GLXContextID = uint;
alias GLXPixmap = uint;
alias GLXDrawable = uint;
alias GLXPbuffer = uint;
alias GLXWindow = uint;
alias GLXFBConfigID = uint;
alias GLXVideoCaptureDeviceNV = XID;
alias GLXPbufferSGIX = XID;
alias GLXVideoSourceSGIX = XID;
alias GLXVideoDeviceNV = uint;


extern(System) {
    alias __GLXextFuncPtr = void function();

    struct GLXPbufferClobberEvent {
        int event_type;             /* GLX_DAMAGED or GLX_SAVED */
        int draw_type;              /* GLX_WINDOW or GLX_PBUFFER */
        ulong serial;       /* # of last request processed by server */
        Bool send_event;            /* true if this came for SendEvent request */
        Display *display;           /* display the event was read from */
        GLXDrawable drawable;       /* XID of Drawable */
        uint buffer_mask;   /* mask indicating which buffers are affected */
        uint aux_buffer;    /* which aux buffer was affected */
        int x, y;
        int width, height;
        int count;                  /* if nonzero, at least this many more */
    }

    struct GLXBufferSwapComplete {
        int type;
        ulong serial;       /* # of last request processed by server */
        Bool send_event;            /* true if this came from a SendEvent request */
        Display *display;           /* Display the event was read from */
        GLXDrawable drawable;       /* drawable on which event was requested in event mask */
        int event_type;
        long ust;
        long msc;
        long sbc;
    }

    union GLXEvent {
        GLXPbufferClobberEvent glxpbufferclobber;
        GLXBufferSwapComplete glxbufferswapcomplete;
        int[24] pad;
    }

    struct GLXBufferClobberEventSGIX {
        int type;
        ulong serial;   /* # of last request processed by server */
        Bool send_event;        /* true if this came for SendEvent request */
        Display *display;       /* display the event was read from */
        GLXDrawable drawable;   /* i.d. of Drawable */
        int event_type;         /* GLX_DAMAGED_SGIX or GLX_SAVED_SGIX */
        int draw_type;          /* GLX_WINDOW_SGIX or GLX_PBUFFER_SGIX */
        uint mask;      /* mask indicating which buffers are affected*/
        int x, y;
        int width, height;
        int count;              /* if nonzero, at least this many more */
    }

    struct GLXHyperpipeNetworkSGIX {
        char[80] pipeName; /* Should be [GLX_HYPERPIPE_PIPE_NAME_LENGTH_SGIX] */
        int      networkId;
    }

    struct GLXHyperpipeConfigSGIX {
        char[80] pipeName; /* Should be [GLX_HYPERPIPE_PIPE_NAME_LENGTH_SGIX] */
        int      channel;
        uint     participationType;
        int      timeSlice;
    }

    struct GLXPipeRect {
        char[80] pipeName; /* Should be [GLX_HYPERPIPE_PIPE_NAME_LENGTH_SGIX] */
        int srcXOrigin, srcYOrigin, srcWidth, srcHeight;
        int destXOrigin, destYOrigin, destWidth, destHeight;
    }

    struct GLXPipeRectLimits {
        char[80] pipeName; /* Should be [GLX_HYPERPIPE_PIPE_NAME_LENGTH_SGIX] */
        int XOrigin, YOrigin, maxHeight, maxWidth;
    }
}
''')
    gen.write_extern(f)
    gen.write_opaque_struct(f, '__GLXcontextRec')
    gen.write_alias(f, 'GLXContext', '__GLXcontextRec*')
    gen.write_opaque_struct(f, '__GLXFBConfigRec')
    gen.write_alias(f, 'GLXFBConfig', '__GLXFBConfigRec*')
    gen.write_alias(f, 'GLXFBConfigSGIX', '__GLXFBConfigRec*')
    gen.write_extern_end(f)


def _wgl_types(gen, f):
    f.write('''
version(Windows) {
    public import core.sys.windows.windows;
} else {
    alias BOOL = int;
    alias CHAR = char;
    alias WORD = ushort;
    alias DWORD = uint;
    alias FLOAT = float;
    alias HANDLE = void*;
    alias HDC = HANDLE;
    alias HGLRC = HANDLE;
    alias INT = int;
    alias LPCSTR = const(CHAR)*;
    alias LPVOID = void*;
    alias UINT = uint;
    alias USHORT = ushort;
    alias VOID = void;
    alias COLORREF = DWORD;
    alias HENHMETAFILE = HANDLE;
    alias BYTE = byte;
}

alias PROC = HANDLE;

extern(System) {
    struct RECT {
        int left;
        int top;
        int right;
        int bottom;
    }

    struct LAYERPLANEDESCRIPTOR {
        WORD     nSize;
        WORD     nVersion;
        DWORD    dwFlags;
        BYTE     iPixelType;
        BYTE     cColorBits;
        BYTE     cRedBits;
        BYTE     cRedShift;
        BYTE     cGreenBits;
        BYTE     cGreenShift;
        BYTE     cBlueBits;
        BYTE     cBlueShift;
        BYTE     cAlphaBits;
        BYTE     cAlphaShift;
        BYTE     cAccumBits;
        BYTE     cAccumRedBits;
        BYTE     cAccumGreenBits;
        BYTE     cAccumBlueBits;
        BYTE     cAccumAlphaBits;
        BYTE     cDepthBits;
        BYTE     cStencilBits;
        BYTE     cAuxBuffers;
        BYTE     iLayerType;
        BYTE     bReserved;
        COLORREF crTransparent;
    }

    struct PIXELFORMATDESCRIPTOR {
        WORD  nSize;
        WORD  nVersion;
        DWORD dwFlags;
        BYTE  iPixelType;
        BYTE  cColorBits;
        BYTE  cRedBits;
        BYTE  cRedShift;
        BYTE  cGreenBits;
        BYTE  cGreenShift;
        BYTE  cBlueBits;
        BYTE  cBlueShift;
        BYTE  cAlphaBits;
        BYTE  cAlphaShift;
        BYTE  cAccumBits;
        BYTE  cAccumRedBits;
        BYTE  cAccumGreenBits;
        BYTE  cAccumBlueBits;
        BYTE  cAccumAlphaBits;
        BYTE  cDepthBits;
        BYTE  cStencilBits;
        BYTE  cAuxBuffers;
        BYTE  iLayerType;
        BYTE  bReserved;
        DWORD dwLayerMask;
        DWORD dwVisibleMask;
        DWORD dwDamageMask;
    }

    struct POINTFLOAT {
        FLOAT x;
        FLOAT y;
    }

    struct GLYPHMETRICSFLOAT {
        FLOAT      gmfBlackBoxX;
        FLOAT      gmfBlackBoxY;
        POINTFLOAT gmfptGlyphOrigin;
        FLOAT      gmfCellIncX;
        FLOAT      gmfCellIncY;
    }
    alias PGLYPHMETRICSFLOAT = GLYPHMETRICSFLOAT*;
    alias LPGLYPHMETRICSFLOAT = GLYPHMETRICSFLOAT;

    struct GPU_DEVICE {
        DWORD      cb;
        CHAR[32]   DeviceName;
        CHAR[128]  DeviceString;
        DWORD      Flags;
        RECT       rcVirtualScreen;
    }

    alias PGPU_DEVICE = GPU_DEVICE;
}
''')
    gen.write_opaque_struct(f, 'HPBUFFERARB')
    gen.write_opaque_struct(f, 'HPBUFFEREXT')
    gen.write_opaque_struct(f, 'HVIDEOOUTPUTDEVICENV')
    gen.write_opaque_struct(f, 'HPVIDEODEV')
    gen.write_opaque_struct(f, 'HPGPUNV')
    gen.write_opaque_struct(f, 'HGPUNV')
    gen.write_opaque_struct(f, 'HVIDEOINPUTDEVICENV')


DTYPES = {
    '__pre': {
        'egl': 'import core.stdc.stdint : intptr_t;\n\n'
    },

    '__other': {
        'gl': _gl_types,
        'egl': _egl_types,
        'glx': _glx_types,
        'wgl': _wgl_types
    },

    'gl': {
        'GLenum': 'uint', 'GLvoid': 'void', 'GLboolean': 'ubyte',
        'GLbitfield': 'uint', 'GLchar': 'char', 'GLbyte': 'byte',
        'GLshort': 'short', 'GLint': 'int', 'GLclampx': 'int',
        'GLsizei': 'int', 'GLubyte': 'ubyte', 'GLushort': 'ushort',
        'GLuint': 'uint', 'GLhalf': 'ushort', 'GLfloat': 'float',
        'GLclampf': 'float', 'GLdouble': 'double', 'GLclampd': 'double',
        'GLfixed': 'int', 'GLintptr': 'ptrdiff_t', 'GLsizeiptr': 'ptrdiff_t',
        'GLintptrARB': 'ptrdiff_t', 'GLsizeiptrARB': 'ptrdiff_t',
        'GLcharARB': 'byte', 'GLhandleARB': 'uint', 'GLhalfARB': 'ushort',
        'GLhalfNV': 'ushort', 'GLint64EXT': 'long', 'GLuint64EXT': 'ulong',
        'GLint64': 'long', 'GLuint64': 'ulong',
        'GLvdpauSurfaceNV': 'ptrdiff_t', 'GLeglImageOES': 'void*'
    },
    'egl': {
        'EGLBoolean': 'uint', 'EGLenum': 'uint', 'EGLAttribKHR': 'intptr_t',
        'EGLAttrib': 'intptr_t', 'EGLClientBuffer': 'void*', 'EGLConfig': 'void*',
        'EGLContext': 'void*', 'EGLDeviceEXT': 'void*', 'EGLDisplay': 'void*',
        'EGLImage': 'void*', 'EGLImageKHR': 'void*', 'EGLOutputLayerEXT': 'void*',
        'EGLOutputPortEXT': 'void*', 'EGLStreamKHR': 'void*', 'EGLSurface': 'void*',
        'EGLSync': 'void*', 'EGLSyncKHR': 'void*', 'EGLSyncNV': 'void*',
        '__eglMustCastToProperFunctionPointerType': 'void function()',
        'EGLint': 'int', 'EGLTimeKHR': 'ulong', 'EGLTime': 'ulong',
        'EGLTimeNV': 'ulong', 'EGLuint64NV': 'ulong',
        'EGLuint64KHR': 'ulong', 'EGLuint64MESA': 'ulong',
        'EGLsizeiANDROID': 'ptrdiff_t', 'EGLNativeFileDescriptorKHR': 'int'
    },
    'glx': {
        'GLboolean': 'ubyte', 'GLenum': 'uint', 'GLint': 'int',
        'GLsizei': 'int', 'GLubyte': 'ubyte', 'GLuint': 'uint',
        'GLfloat': 'float', 'GLbitfield': 'uint', 'GLintptr': 'ptrdiff_t',
        'GLsizeiptr': 'ptrdiff_t'

    },
    'wgl': {
        'GLbitfield': 'uint', 'GLenum': 'uint', 'GLfloat': 'float',
        'GLint': 'int', 'GLsizei': 'int', 'GLuint': 'uint',
        'GLushort': 'ushort', 'INT32': 'int', 'INT64': 'long',
        'GLboolean': 'ubyte'
    },

    'SpecialNumbers': {
        'gl': [
            ('GL_FALSE', '0', 'ubyte'), ('GL_TRUE', '1', 'ubyte'),
            ('GL_NO_ERROR', '0', 'uint'), ('GL_NONE', '0', 'uint'),
            ('GL_ZERO', '0', 'uint'), ('GL_ONE', '1', 'uint'),
            ('GL_NONE_OES', '0', 'uint'),
            ('GL_INVALID_INDEX', '0xFFFFFFFF', 'uint'),
            ('GL_TIMEOUT_IGNORED', '0xFFFFFFFFFFFFFFFF', 'ulong'),
            ('GL_TIMEOUT_IGNORED_APPLE', '0xFFFFFFFFFFFFFFFF', 'ulong'),
            ('GL_VERSION_ES_CL_1_0', '1', 'uint'), ('GL_VERSION_ES_CM_1_1', '1', 'uint'),
            ('GL_VERSION_ES_CL_1_1', '1', 'uint')
        ],
        'egl': [
            ('EGL_DONT_CARE', '-1', 'int'), ('EGL_UNKNOWN', '-1', 'int'),
            ('EGL_NO_NATIVE_FENCE_FD_ANDROID', '-1', 'uint'),
            ('EGL_DEPTH_ENCODING_NONE_NV', '0', 'uint'),
            ('EGL_NO_CONTEXT', 'cast(EGLContext)0', 'EGLContext'),
            ('EGL_NO_DEVICE_EXT', 'cast(EGLDeviceEXT)0', 'EGLDeviceEXT'),
            ('EGL_NO_DISPLAY', 'cast(EGLDisplay)0', 'EGLDisplay'),
            ('EGL_NO_IMAGE', 'cast(EGLImage)0', 'EGLImage'),
            ('EGL_NO_IMAGE_KHR', 'cast(EGLImageKHR)0', 'EGLImageKHR'),
            ('EGL_DEFAULT_DISPLAY', 'cast(EGLNativeDisplayType)0', 'EGLNativeDisplayType'),
            ('EGL_NO_FILE_DESCRIPTOR_KHR', 'cast(EGLNativeFileDescriptorKHR)-1', 'EGLNativeFileDescriptorKHR'),
            ('EGL_NO_OUTPUT_LAYER_EXT', 'cast(EGLOutputLayerEXT)0', 'EGLOutputLayerEXT'),
            ('EGL_NO_OUTPUT_PORT_EXT', 'cast(EGLOutputPortEXT)0', 'EGLOutputPortEXT'),
            ('EGL_NO_STREAM_KHR', 'cast(EGLStreamKHR)0', 'EGLStreamKHR'),
            ('EGL_NO_SURFACE', 'cast(EGLSurface)0', 'EGLSurface'),
            ('EGL_NO_SYNC', 'cast(EGLSync)0', 'EGLSync'),
            ('EGL_NO_SYNC_KHR', 'cast(EGLSyncKHR)0', 'EGLSyncKHR'),
            ('EGL_NO_SYNC_NV', 'cast(EGLSyncNV)0', 'EGLSyncNV'),
            ('EGL_DISPLAY_SCALING', '10000', 'uint'),
            ('EGL_FOREVER', '0xFFFFFFFFFFFFFFFF', 'ulong'),
            ('EGL_FOREVER_KHR', '0xFFFFFFFFFFFFFFFF', 'ulong'),
            ('EGL_FOREVER_NV', '0xFFFFFFFFFFFFFFFF', 'ulong')
        ],
        'glx': [
            ('GLX_DONT_CARE', '0xFFFFFFFF', 'uint'),
            ('GLX_CONTEXT_RELEASE_BEHAVIOR_NONE_ARB', '0', 'uint')
        ],
        'wgl': [
            ('WGL_CONTEXT_RELEASE_BEHAVIOR_NONE_ARB', '0', 'uint'),
            ('WGL_FONT_LINES', '0', 'uint'),
            ('WGL_FONT_POLYGONS', 1, 'uint')
        ]
    }
}


class BaseDGenerator(Generator):
    NAME = 'd'
    NAME_LONG = 'D'

    def open(self):
        self._f_loader = open(self.make_path(self.LOADER), 'w')
        self._f_gl = open(self.make_path(self.PACKAGE), 'w')
        self._f_types = open(self.make_path(self.TYPES), 'w')
        self._f_enums = open(self.make_path(self.ENUMS), 'w')
        self._f_funcs = open(self.make_path(self.FUNCS), 'w')
        self._f_exts = open(self.make_path(self.EXT), 'w')

    def close(self):
        self._f_loader.close()
        self._f_gl.close()
        self._f_types.close()
        self._f_enums.close()
        self._f_funcs.close()
        self._f_exts.close()

    @property
    def PACKAGE(self):
        return 'all'

    def generate_header(self):
        self._f_gl.write('/*\n')
        self._f_gl.write(self.header)
        self._f_gl.write('*/\n\n')

    def generate_loader(self, features, extensions):
        f = self._f_loader

        rfeatures = features
        if self.spec.NAME in ('egl', 'wgl'):
            features = {'egl': [], 'wgl': []}

        self.write_module(f, self.LOADER)
        self.write_imports(f, [self.FUNCS, self.EXT, self.ENUMS, self.TYPES])

        self.loader.write(f)
        self.loader.write_has_ext(f)

        written = set()
        for api, version in self.api.items():
            loadername = 'Load' if self.LOAD_GL_PREFIX else 'load'
            f.write('bool {}{}{}(Loader load) {{\n'
                    .format(self.LOAD_GL_PREFIX, loadername, api.upper()))
            self.loader.write_begin_load(f)
            f.write('\tfind_core{}();\n'.format(api.upper()))
            for feature in features[api]:
                f.write('\tload_{}(load);\n'.format(feature.name))
            f.write('\n\tfind_extensions{}();\n'.format(api.upper()))
            for ext in extensions[api]:
                if len(list(ext.functions)) == 0:
                    continue
                f.write('\tload_{}(load);\n'.format(ext.name))
            self.loader.write_end_load(f)
            f.write('}\n\n')

            f.write('private {\n\n')

            f.write('void find_core{}() {{\n'.format(api.upper()))
            self.loader.write_find_core(f)
            if self.spec.NAME == 'gl':
                for feature in features[api]:
                    f.write('\t{} = (major == {num[0]} && minor >= {num[1]}) ||'
                        ' major > {num[0]};\n'.format(feature.name, num=feature.number))
            f.write('\treturn;\n')
            f.write('}\n\n')

            f.write('void find_extensions{}() {{\n'.format(api.upper()))
            if self.spec.NAME == 'gl':
                for ext in extensions[api]:
                    f.write('\t{0} = has_ext("{0}");\n'.format(ext.name))
            f.write('\treturn;\n')
            f.write('}\n\n')

            for feature in features[api]:
                f.write('void load_{}(Loader load) {{\n'
                            .format(feature.name))
                if self.spec.NAME == 'gl':
                    f.write('\tif(!{}) return;\n'.format(feature.name))
                for func in feature.functions:
                    f.write('\t{name} = cast(typeof({name}))load("{name}");\n'
                        .format(name=func.proto.name))
                f.write('\treturn;\n}\n\n')

            for ext in extensions[api]:
                if len(list(ext.functions)) == 0 or ext.name in written:
                    continue

                f.write('void load_{}(Loader load) {{\n'
                    .format(ext.name))
                if self.spec.NAME == 'gl':
                    f.write('\tif(!{}) return;\n'.format(ext.name))
                for func in ext.functions:
                    # even if they were in written we need to load it
                    f.write('\t{name} = cast(typeof({name}))load("{name}");\n'
                        .format(name=func.proto.name))
                f.write('\treturn;\n')
                f.write('}\n')

                written.add(ext.name)

            f.write('\n} /* private */\n\n')

        self.write_packages(rfeatures, extensions)

    def write_packages(self, allfeatures, allextensions):
        f = self._f_gl

        self.write_module(f, self.PACKAGE)
        self.write_imports(f, [self.FUNCS, self.EXT, self.ENUMS, self.TYPES], False)

        for api, features in allfeatures.items():
            extensions = allextensions[api]
            with open(self.make_path(api), 'w') as f:
                self.write_module(f, api)

                self.write_imports(f, [self.TYPES], False)

                extenums = chain.from_iterable(ext.enums for ext in extensions)
                funcenums = chain.from_iterable(ext.enums for ext in extensions)
                enums = set(enum.name for enum in extenums) | \
                        set(enum.name for enum in funcenums)

                featfuncs = set(func.proto.name for func in
                        chain.from_iterable(feat.functions for feat in features))
                extfuncs = set(func.proto.name for func in
                        chain.from_iterable(ext.functions for ext in extensions))
                extfuncs = extfuncs - featfuncs

                enums |= set(enum.name for enum in
                        chain.from_iterable(feat.enums for feat in features))

                self.write_selective_import(f, self.FUNCS, featfuncs)
                self.write_selective_import(f, self.EXT, extfuncs)
                self.write_selective_import(f, self.ENUMS, enums)


    def generate_types(self, types):
        f = self._f_types

        self.write_module(f, self.TYPES)

        f.write(self.TYPE_DICT.get('__pre', {}).get(self.spec.NAME,''))
        for ogl, d in self.TYPE_DICT[self.spec.NAME].items():
            self.write_alias(f, ogl, d)
        self.TYPE_DICT['__other'][self.spec.NAME](self, f)

    def generate_features(self, features):
        self.write_enums(features)
        self.write_funcs(features)

    def write_enums(self, features):
        e = self._f_enums

        self.write_module(e, self.ENUMS)
        self.write_imports(e, [self.TYPES])

        for v in self.TYPE_DICT['SpecialNumbers'][self.spec.NAME]:
            self.write_enum(e, *v)

        written = set()
        for feature in features:
            for enum in feature.enums:
                if enum.group == 'SpecialNumbers':
                    written.add(enum)
                    continue
                if not enum in written:
                    self.write_enum(e, enum.name, enum.value)
                written.add(enum)

    def write_funcs(self, features):
        f = self._f_funcs

        self.write_module(f, self.FUNCS)
        self.write_imports(f, [self.TYPES])

        if self.spec.NAME == 'gl':
            for feature in features:
                self.write_boolean(f, feature.name)

        if self.spec.NAME in ('egl', 'wgl'):
            self.write_extern(f)
            for feature in features:
                for func in feature.functions:
                    self.write_function_def(f, func)
            self.write_extern_end(f)
        else:
            self.write_functions(f, set(), set(), features)

    def generate_extensions(self, extensions, enums, functions):
        f = self._f_exts

        self.write_module(f, self.EXT)
        self.write_imports(f, [self.TYPES, self.ENUMS, self.FUNCS])

        write = set()
        written = set(enum.name for enum in enums) | \
                    set(function.proto.name for function in functions)
        for ext in extensions:
            if self.spec.NAME == 'gl' and not ext.name in written:
                self.write_boolean(f, ext.name)
            for enum in ext.enums:
                if not enum.name in written and not enum.group == 'SpecialNumbers':
                    self.write_enum(self._f_enums, enum.name, enum.value)
                written.add(enum.name)
            written.add(ext.name)

        self.write_functions(f, write, written, extensions)

    def write_functions(self, f, write, written, extensions):
        self.write_prototype_pre(f)
        for ext in extensions:
            for func in ext.functions:
                if not func.proto.name in written:
                    self.write_function_prototype(f, func)
                    write.add(func)
                written.add(func.proto.name)
        self.write_prototype_post(f)

        self.write_function_pre(f)
        for func in write:
            self.write_function(f, func)
        self.write_function_post(f)


    def make_path(self, name):
        path = os.path.join(self.path, self.MODULE.split('.')[-1],
                            self.spec.NAME, name.split('.')[-1] + self.FILE_EXTENSION)
        makefiledir(path)
        return path

    def write_imports(self, fobj, modules, private=True):
        raise NotImplementedError

    def write_selective_import(self, fobj, imports):
        raise NotImplementedError

    def write_module(self, fobj, name):
        raise NotImplementedError

    def write_prototype_pre(self, fobj):
        raise NotImplementedError

    def write_prototype_post(self, fobj):
        raise NotImplementedError

    def write_function_pre(self, fobj):
        raise NotImplementedError

    def write_function_post(self, fobj):
        raise NotImplementedError

    def write_extern(self, fobj):
        raise NotImplementedError

    def write_extern_end(self, fobj):
        raise NotImplementedError

    def write_shared(self, fobj):
        raise NotImplementedError

    def write_shared_end(self, fobj):
        raise NotImplementedError

    def write_function_def(self, fobj, func):
        raise NotImplementedError

    def write_function(self, fobj, func):
        raise NotImplementedError

    def write_function_prototype(self, fobj, func):
        raise NotImplementedError

    def write_boolean(self, fobj, name, value=False):
        raise NotImplementedError

    def write_enum(self, fobj, name, value, type='uint'):
        raise NotImplementedError

    def write_opaque_struct(self, fobj, name):
        raise NotImplementedError

    def write_alias(self, fobj, newn, decl):
        raise NotImplementedError


class DGenerator(BaseDGenerator):
    MODULE = 'glad'
    LOADER = 'loader'
    ENUMS = 'enums'
    EXT = 'ext'
    FUNCS = 'funcs'
    TYPES = 'types'
    FILE_EXTENSION = '.d'
    TYPE_DICT = DTYPES

    LOAD_GL_PREFIX = 'glad'

    def write_imports(self, fobj, modules, private=True):
        for mod in modules:
            if private:
                fobj.write('private ')
            else:
                fobj.write('public ')

            fobj.write('import {}.{}.{};\n'.format(self.MODULE, self.spec.NAME, mod))

    def write_selective_import(self, fobj, mod, imports):
        if len(imports) == 0: return

        fobj.write('public import {}.{}.{} :\n'.format(self.MODULE, self.spec.NAME, mod))
        imports = set(imports)
        last = len(imports)
        for i, im in enumerate(imports, 1):
            fobj.write(im)
            if not i == last:
                fobj.write(', ')
            if (i % 5) == 0:
                fobj.write('\n')
        fobj.write(';\n\n')

    def write_module(self, fobj, name):
        fobj.write('module {}.{}.{};\n\n\n'.format(self.MODULE, self.spec.NAME, name))

    def write_prototype_pre(self, fobj):
        fobj.write('nothrow @nogc ')
        self.write_extern(fobj)

    def write_prototype_post(self, fobj):
        self.write_extern_end(fobj)

    def write_function_pre(self, fobj):
        self.write_shared(fobj)

    def write_function_post(self, fobj):
        self.write_shared_end(fobj)

    def write_extern(self, fobj):
        fobj.write('extern(System) {\n')

    def write_extern_end(self, fobj):
        fobj.write('}\n')

    def write_shared(self, fobj):
        fobj.write('__gshared {\n')

    def write_shared_end(self, fobj):
        fobj.write('}\n')

    def write_function_def(self, fobj, func):
        fobj.write('{} {}('.format(func.proto.ret.to_d(), func.proto.name))
        fobj.write(', '.join(param.type.to_d() for param in func.params))
        fobj.write(');\n')

    def write_function(self, fobj, func):
        fobj.write('fp_{0} {0};\n'.format(func.proto.name))

    def write_function_prototype(self, fobj, func):
        fobj.write('alias fp_{} = {} function('
                .format(func.proto.name, func.proto.ret.to_d()))
        fobj.write(', '.join(param.type.to_d() for param in func.params))
        fobj.write(');\n')

    def write_boolean(self, fobj, name, value=False):
        if value:
            fobj.write('bool {} = true;\n'.format(name))
        else:
            fobj.write('bool {};\n'.format(name))

    def write_enum(self, fobj, name, value, type='uint'):
        if isinstance(value, basestring) and '"' in value:
            type = 'const(char)*'

        fobj.write('enum {} {} = {};\n'.format(type, name, value))

    def write_opaque_struct(self, fobj, name):
        fobj.write('struct _{name}; alias {name} = _{name}*;\n'.format(name=name))

    def write_alias(self, fobj, newn, decl):
        fobj.write('alias {} = {};\n'.format(newn, decl))




import glad.lang.c
import glad.lang.d
import glad.lang.nim
import glad.lang.pascal
import glad.lang.volt


def get_generator(name, spec):
    _langs = [
        glad.lang.c,
        glad.lang.d,
        glad.lang.nim,
        glad.lang.pascal,
        glad.lang.volt
    ]

    for lang in _langs:
        gen, loader = lang.get_generator(name, spec)
        if gen is not None:
            return gen, loader
    return None, None



from glad.lang.common.loader import BaseLoader
from glad.lang.c.loader import LOAD_OPENGL_DLL, LOAD_OPENGL_DLL_H, LOAD_OPENGL_GLAPI_H

_WGL_LOADER = \
    LOAD_OPENGL_DLL % {'pre':'static', 'init':'open_gl',
                       'proc':'get_proc', 'terminate':'close_gl'} + '''
int gladLoadWGL(HDC hdc) {
    int status = 0;

    if(open_gl()) {
        status = gladLoadWGLLoader((GLADloadproc)get_proc, hdc);
        close_gl();
    }

    return status;
}
'''

_WGL_HEADER_START = '''
#ifndef WINAPI
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN 1
#endif
#ifndef NOMINMAX
#define NOMINMAX 1
#endif
#include <windows.h>
#endif

'''

#include <glad/glad.h>

_WGL_HEADER_MID = '''
#ifndef __glad_wglext_h_

#ifdef __wglext_h_
#error WGL header already included, remove this include, glad already provides it
#endif

#define __glad_wglext_h_
#define __wglext_h_

#ifndef APIENTRY
#define APIENTRY
#endif
#ifndef APIENTRYP
#define APIENTRYP APIENTRY *
#endif

#ifdef __cplusplus
extern "C" {
#endif

typedef void* (* GLADloadproc)(const char *name);
''' + LOAD_OPENGL_GLAPI_H

_WGL_HEADER_LOADER = '''
GLAPI int gladLoadWGL(HDC hdc);
''' + LOAD_OPENGL_DLL_H

_WGL_HEADER_END = '''
#ifdef __cplusplus
}
#endif

#endif
'''

_WGL_HAS_EXT = '''
static HDC GLADWGLhdc = (HDC)INVALID_HANDLE_VALUE;

static int get_exts(void) {
    return 1;
}

static void free_exts(void) {
    return;
}

static int has_ext(const char *ext) {
    const char *terminator;
    const char *loc;
    const char *extensions;

    if(wglGetExtensionsStringEXT == NULL && wglGetExtensionsStringARB == NULL)
        return 0;

    if(wglGetExtensionsStringARB == NULL || GLADWGLhdc == INVALID_HANDLE_VALUE)
        extensions = wglGetExtensionsStringEXT();
    else
        extensions = wglGetExtensionsStringARB(GLADWGLhdc);

    if(extensions == NULL || ext == NULL)
        return 0;

    while(1) {
        loc = strstr(extensions, ext);
        if(loc == NULL)
            break;

        terminator = loc + strlen(ext);
        if((loc == extensions || *(loc - 1) == ' ') &&
            (*terminator == ' ' || *terminator == '\\0'))
        {
            return 1;
        }
        extensions = terminator;
    }

    return 0;
}
'''

class WGLCLoader(BaseLoader):
    def write(self, fobj):
        if not self.disabled:
            fobj.write(_WGL_LOADER)

    def write_begin_load(self, fobj):
        fobj.write('\twglGetExtensionsStringARB = (PFNWGLGETEXTENSIONSSTRINGARBPROC)load("wglGetExtensionsStringARB");\n')
        fobj.write('\twglGetExtensionsStringEXT = (PFNWGLGETEXTENSIONSSTRINGEXTPROC)load("wglGetExtensionsStringEXT");\n')
        fobj.write('\tif(wglGetExtensionsStringARB == NULL && wglGetExtensionsStringEXT == NULL) return 0;\n')

    def write_end_load(self, fobj):
        fobj.write('\treturn 1;\n')

    def write_find_core(self, fobj):
        fobj.write('\tGLADWGLhdc = hdc;\n')

    def write_has_ext(self, fobj):
        fobj.write(_WGL_HAS_EXT)

    def write_header(self, fobj):
        fobj.write(_WGL_HEADER_START)
        if self.local_files:
            fobj.write('#include "glad.h"\n')
        else:
            fobj.write('#include <glad/glad.h>\n')
        fobj.write(_WGL_HEADER_MID)
        if not self.disabled:
            fobj.write(_WGL_HEADER_LOADER)

    def write_header_end(self, fobj):
        fobj.write(_WGL_HEADER_END)
